import os
import re
from typing import List

from gimie.graph.namespaces import GIMIE


def locate_licenses(path: str, recurse: bool = False) -> List[str]:
    """Returns valid potential paths to license files in the project.
    This uses pattern-matching on file names.

    Parameters
    ----------
    path:
        The root path to search for license files.
    recurse:
        Whether to look into subdirectories.

    Returns
    -------
    license_files:
        The list of relative paths (from input path) to files which
        potentially contain license text.

    Examples
    --------
    >>> locate_licenses('.')
    ['./LICENSE']
    """
    license_files = []
    pattern = r".*(license(s)?|reus(e|ing)|copy(ing)?)(\.(txt|md|rst))?$"

    for root, _, files in os.walk(path):
        # skip toplevel hidden dirs (e.g. .git/)
        subdir = os.path.relpath(root, path)
        if subdir.startswith(".") and subdir != ".":
            continue
        for file in files:
            # skip hidden files
            if file.startswith("."):
                continue

            if re.match(pattern, file, flags=re.IGNORECASE):
                license_path = os.path.join(root, file)
                license_files.append(license_path)

        # The first root of os.walk is the current dir
        if not recurse:
            return license_files

    return license_files


def get_spdx_url(name: str) -> str:
    """Given an SPDX license identifier, return the full URL."""
    return f"https://spdx.org/licenses/{name}"



import requests
from typing import Any, Dict, List, Union


def send_rest_query(
    api: str, query: str, headers: Dict[str, str]
) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    """Generic function to send a query to the GitHub/GitLab rest API."""
    resp = requests.get(
        url=f"{api}/{query}",
        headers=headers,
    )

    if resp.status_code != 200:
        raise ConnectionError(resp.json()["message"])
    return resp.json()


def send_graphql_query(
    api: str, query: str, data: Dict[str, Any], headers: Dict[str, str]
) -> Dict[str, Any]:
    """Generic function to send a GraphQL query to the GitHub/GitLab API."""
    resp = requests.post(
        url=f"{api}/graphql",
        json={
            "query": query,
            "variables": data,
        },
        headers=headers,
    )

    if resp.status_code != 200:
        raise ConnectionError(resp.json()["message"])
    return resp.json()



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extractor which uses a locally available (usually cloned) repository."""
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Union
import uuid

from calamus import fields
from calamus.schema import JsonLDSchema
import git
import pydriller
from rdflib import Graph

from gimie.models import Person, PersonSchema
from gimie.graph.namespaces import SDO
from gimie.sources.abstract import Extractor
from gimie.utils import generate_uri


@dataclass
class GitExtractor(Extractor):
    """
    This class is responsible for extracting metadata from a git repository.

    Parameters
    ----------
    path: str
        The path to the git repository.

    Attributes
    ----------
    uri: Optional[str]
        The URI to assign the repository in RDF.
    repository: Repository
        The repository we are extracting metadata from.
    """

    path: str
    _id: Optional[str] = None
    author: Optional[Person] = None
    contributors: Optional[List[Person]] = None
    date_created: Optional[datetime] = None
    date_modified: Optional[datetime] = None

    def extract(self):
        self.repository = pydriller.Repository(self.path)
        if self._id is None:
            head_commit_hash = git.Repo(self.path).head.commit.hexsha[:7]
            self._id = generate_uri(head_commit_hash)
        # Assuming author is the first person to commit
        self.author = self._get_creator()
        self.contributors = self._get_contributors()
        self.date_created = self._get_creation_date()
        self.date_modified = self._get_modification_date()

    def to_graph(self) -> Graph:
        """Generate an RDF graph from the instance"""
        jd = GitExtractorSchema().dumps(self)
        g: Graph = Graph().parse(data=str(jd), format="json-ld")
        g.bind("schema", SDO)
        return g

    def _get_contributors(self) -> List[Person]:
        """Get the authors of the repository."""
        authors = set()
        for commit in self.repository.traverse_commits():
            if commit.author is not None:
                authors.add((commit.author.name, commit.author.email))
        return [self._dev_to_person(name, email) for name, email in authors]

    def _get_creation_date(self) -> Optional[datetime]:
        """Get the creation date of the repository."""
        try:
            return next(self.repository.traverse_commits()).author_date
        except StopIteration:
            return None

    def _get_modification_date(self) -> Optional[datetime]:
        """Get the last modification date of the repository."""
        try:
            for commit in self.repository.traverse_commits():
                pass
            return commit.author_date
        except (StopIteration, NameError):
            return None

    def _get_creator(self) -> Optional[Person]:
        """Get the creator of the repository."""
        try:
            creator = next(self.repository.traverse_commits()).author
            return self._dev_to_person(creator.name, creator.email)
        except StopIteration:
            return None

    def _dev_to_person(
        self, name: Optional[str], email: Optional[str]
    ) -> Person:
        """Convert a Developer object to a Person object."""
        if name is None:
            uid = str(uuid.uuid4())
        else:
            uid = name.replace(" ", "_").lower()
        dev_id = f"{self._id}/{uid}"
        return Person(
            _id=dev_id,
            identifier=uid,
            name=name,
            email=email,
        )


class GitExtractorSchema(JsonLDSchema):
    _id = fields.Id()
    author = fields.Nested(SDO.author, PersonSchema)
    contributors = fields.Nested(SDO.contributor, PersonSchema, many=True)
    date_created = fields.Date(SDO.dateCreated)
    date_modified = fields.Date(SDO.dateModified)

    class Meta:
        rdf_type = SDO.SoftwareSourceCode
        model = GitExtractor
        add_value_types = False



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sources from which metadata can be extracted by gimie."""
from typing import Type
from gimie.sources.abstract import Extractor
from gimie.sources.github import GithubExtractor
from gimie.sources.gitlab import GitlabExtractor
from gimie.sources.git import GitExtractor

from dataclasses import dataclass


@dataclass
class Source:
    """Source of metadata."""

    remote: bool
    git: bool
    extractor: Type[Extractor]


SOURCES = {
    "git": Source(remote=False, git=True, extractor=GitExtractor),
    "github": Source(remote=True, git=True, extractor=GithubExtractor),
    "gitlab": Source(remote=True, git=True, extractor=GitlabExtractor),
}



from __future__ import annotations
from dataclasses import dataclass
import os
import requests
from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse

from dotenv import load_dotenv
from calamus import fields
from calamus.schema import JsonLDSchema
from rdflib import Graph

from gimie.sources.abstract import Extractor
from gimie.models import (
    Organization,
    OrganizationSchema,
    Person,
    PersonSchema,
)
from gimie.graph.namespaces import SDO
from gimie.sources.common.queries import send_graphql_query, send_rest_query

GL_API_REST = "https://gitlab.com/api/v4/"
GL_API_GRAPHQL = "https://gitlab.com/api"
load_dotenv()


@dataclass
class GitlabExtractor(Extractor):
    """Extractor for Gitlab repositories. Uses the Gitlab GraphQL API to
    extract metadata into linked data."""

    path: str
    _id: Optional[str] = None
    token: Optional[str] = None

    name: Optional[str] = None
    identifier: Optional[str] = None
    author: Optional[List[Union[Organization, Person]]] = None
    contributors: Optional[List[Person]] = None
    prog_langs: Optional[List[str]] = None
    description: Optional[str] = None
    date_created: Optional[datetime] = None
    date_modified: Optional[datetime] = None
    version: Optional[str] = None
    keywords: Optional[List[str]] = None
    source_organization: Optional[Organization] = None
    download_url: Optional[str] = None
    # license: Optional[str] = None

    def to_graph(self) -> Graph:
        """Convert repository to RDF graph."""
        jd = GitlabExtractorSchema().dumps(self)
        g: Graph = Graph().parse(format="json-ld", data=str(jd))
        g.bind("schema", SDO)
        return g

    def extract(self):
        """Extract metadata from target Gitlab repository."""
        if self._id is None:
            self._id = self.path
        self.name = urlparse(self.path).path.strip("/")

        # fetch metadata
        data = self._fetch_repo_data(self.name)

        # Each Gitlab project has a unique identifier (integer)
        self.identifier = urlparse(data["id"]).path.split("/")[2]
        # at the moment, Gimie fetches only the group directly related to the project
        # the group name will take the form: parent/subgroup
        self.source_organization = self._safe_extract_group(data)
        self.description = data["description"]
        self.prog_langs = [lang["name"] for lang in data["languages"]]
        self.date_created = datetime.fromisoformat(data["createdAt"][:-1])
        self.date_modified = datetime.fromisoformat(
            data["lastActivityAt"][:-1]
        )
        self.keywords = data["topics"]

        # Get contributors as the project members that are not owners and those that have written merge requests
        # owners are either multiple individuals or a group, if not user is marked as owner
        self.author = self._safe_extract_author(data)
        # contributors are project members or merge request authors
        self.contributors = self._safe_extract_contributors(data)

        if data["releases"] and (len(data["releases"]["edges"]) > 0):
            # go into releases and take the name from the first node (most recent)
            self.version = data["releases"]["edges"][0]["node"]["name"]
            self.download_url = f"{self.path}/-/archive/{self.version}/{self.name.split('/')[-1]}-{self.version}.tar.gz"

        # for the license, we need to query the rest API
        # the code below does not work, returns - if you have permission- the GitLab specific licence
        # resp = requests.get(url=f"{GL_API_rest}/license/{self.identifier}")
        # if resp.status_code == 200:
        #     self.license = resp.json()

    def _safe_extract_group(
        self, repo: Dict[str, Any]
    ) -> Optional[Organization]:
        """Extract the group from a GraphQL repository node if it has one."""
        if (self.name is not None) and (repo["group"] is not None):
            repo["group"]["name"] = "/".join(self.name.split("/")[0:-1])
            return self._get_organization(repo["group"])
        return None

    def _safe_extract_author(
        self, repo: Dict[str, Any]
    ) -> List[Union[Person, Organization]]:
        """Extract the author from a GraphQL repository node.
        projectMembers is used if available, otherwise the author
        is inferred from the project url."""
        members = repo["projectMembers"]["edges"]
        if len(members) > 0:
            owners = filter(
                lambda m: m["node"]["accessLevel"]["stringValue"] == "OWNER",
                members,
            )
            return [
                self._get_author(owner["node"]["user"]) for owner in owners
            ]

        if repo["group"] is not None:
            return [self._get_author(repo["group"])]

        # If the author is absent from the GraphQL response (permission bug),
        # fallback to the REST API
        return [self._user_from_rest(self.name.split("/")[0])]

    def _safe_extract_contributors(
        self, repo: dict[str, Any]
    ) -> list[Person] | None:
        members = [
            user["node"]["user"]
            for user in repo["projectMembers"]["edges"]
            if user["node"]["accessLevel"]["stringValue"] != "OWNER"
        ]
        merge_request_authors = [
            author["node"]["author"]
            for author in repo["mergeRequests"]["edges"]
        ]
        contributors = members + merge_request_authors
        # Drop duplicate (unhashable) dicts by "id" key
        uniq_contrib = list({c["id"]: c for c in contributors}.values())
        return [self._get_user(contrib) for contrib in uniq_contrib]

    def _fetch_repo_data(self, path: str) -> Dict[str, Any]:
        """Fetch repository metadata from GraphQL endpoint."""
        data = {"path": path}
        project_query = """
        query project_query($path: ID!) {
            project(fullPath: $path) {
                name
                id
                description
                createdAt
                lastActivityAt
                group {
                    id
                    name
                    description
                    avatarUrl
                    webUrl
                }
                languages {
                    name
                    share
                }
                topics
                projectMembers {
                    edges {
                        node {
                        id
                        accessLevel {
                            stringValue
                        }
                        user {
                            id
                            name
                            username
                            publicEmail
                            webUrl
                        }
                        }
                    }
                }
                mergeRequests{
                    edges {
                    node {
                        author {
                        id
                        name
                        username
                        publicEmail
                        webUrl
                        }
                    }
                    }
                }
                releases {
                    edges {
                    node {
                        name
                    }
                    }
                }
        }
        }
        """
        response = send_graphql_query(
            GL_API_GRAPHQL, project_query, data, self._set_auth()
        )
        if "errors" in response:
            raise ValueError(response["errors"])

        return response["data"]["project"]

    def _set_auth(self) -> Any:
        """Set authentication headers for Gitlab API requests."""
        try:
            if not self.token:
                self.token = os.environ.get("GITLAB_TOKEN")
                assert self.token
            headers = {"Authorization": f"token {self.token}"}

            login = requests.get(f"{GL_API_REST}/user", headers=headers)
            assert login.json().get("login")
        except AssertionError:
            return {}
        else:
            return headers

    def _get_author(self, node: Dict[str, Any]) -> Union[Organization, Person]:
        """Given the GraphQL node for a repository owner,
        return the author as a Person or Organization object."""
        # Is this the best test?
        if "username" in node:
            return self._get_user(node)
        return self._get_organization(node)

    def _get_organization(self, node: Dict[str, Any]) -> Organization:
        """Extract details from a GraphQL organization node."""
        return Organization(
            _id=node["webUrl"],
            name=node["name"],
            description=node.get("description"),
            logo=node.get("avatarUrl"),
        )

    def _get_user(self, node: Dict[str, Any]) -> Person:
        """Extract details from a GraphQL user node."""
        return Person(
            _id=node["webUrl"],
            identifier=node["username"],
            name=node.get("name"),
            email=node.get("publicEmail"),
        )

    def _user_from_rest(self, username: str) -> Person:
        """Given a username, use the REST API to retrieve the Person object."""

        author = send_rest_query(
            GL_API_REST,
            f"/users?username={username}",
            self._set_auth(),
        )
        if isinstance(author, list):
            author = author[0]

        return Person(
            _id=author["web_url"],
            identifier=author["username"],
            name=author.get("name"),
        )


class GitlabExtractorSchema(JsonLDSchema):
    """This defines the schema used for json-ld serialization."""

    _id = fields.Id()
    name = fields.String(SDO.name)
    identifier = fields.String(SDO.identifier)
    source_organization = fields.Nested(SDO.isPartOf, OrganizationSchema)
    author = fields.Nested(
        SDO.author, [PersonSchema, OrganizationSchema], many=True
    )
    contributors = fields.Nested(SDO.contributor, PersonSchema, many=True)
    prog_langs = fields.List(SDO.programmingLanguage, fields.String)
    download_url = fields.Raw(SDO.downloadUrl)
    description = fields.String(SDO.description)
    date_created = fields.Date(SDO.dateCreated)
    date_modified = fields.Date(SDO.dateModified)
    # license = IRI(SDO.license)
    path = fields.IRI(SDO.CodeRepository)
    keywords = fields.List(SDO.keywords, fields.String)
    version = fields.String(SDO.version)

    class Meta:
        rdf_type = SDO.SoftwareSourceCode
        model = GitlabExtractor
        add_value_types = False



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from rdflib.namespace import Namespace

SDO = Namespace("http://schema.org/")
COD = Namespace("https://doi.org/10.5063/schema/codemeta-2.0/")
SD = Namespace("https://w3id.org/okn/o/sd/1.9.0/")
BIO = Namespace("https://bioschemas.org/")
GIMIE = Namespace("https://sdsc-ord.github.io/gimie/")



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line interface to the gimie package."""
from enum import Enum
from typing import Optional
from gimie import __version__
import click
import typer

from gimie.project import Project

app = typer.Typer(add_completion=False)

# Used to autogenerate docs with sphinx-click
@click.group()
def cli():
    """Command line group"""
    pass


class SerializationFormat(str, Enum):
    """Enumeration of valid RDF serialization formats for project graphs"""

    ttl = "ttl"
    jsonld = "json-ld"
    nt = "nt"


def version_callback(value: bool):
    if value:
        print(f"gimie {__version__}")
        # Exits successfully
        raise typer.Exit()


@app.command()
def data(
    url: str,
    format: SerializationFormat = typer.Option(
        "ttl",
        "--format",
        show_choices=True,
        help="Output serialization format for the RDF graph.",
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        help="Display version and exit",
        callback=version_callback,
    ),
):
    """Extract linked metadata from a Git repository at the target URL.

    The output is sent to stdout, and turtle is used as the default serialization format."""
    proj = Project(url)
    print(proj.serialize(format=format))


@app.command()
def advice(url: str):
    """Show a metadata completion report for a Git repository
    at the target URL.

    NOTE: Not implemented yet"""
    ...
    raise typer.Exit()


typer_cli = typer.main.get_command(app)
cli.add_command(typer_cli, "cli")

# This callback is triggered when gimie is called without subcommand
@app.callback()
def callback(
    version: Optional[bool] = typer.Option(
        None, "--version", callback=version_callback
    )
):
    """gimie digs Git repositories for metadata."""


if __name__ == "__main__":
    app()



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data models to represent nodes in the graph generated by gimie."""
from __future__ import annotations
from dataclasses import dataclass, field
import datetime
from typing import Optional, List

from calamus.schema import JsonLDSchema
from calamus import fields

from gimie.graph.namespaces import SDO


@dataclass(order=True)
class Release:
    """
    This class represents a release of a repository.

    Parameters
    ----------
    tag: str
        The tag of the release.
    date: datetime.datetime
        The date of the release.
    commit_hash: str
        The commit hash of the release.
    """

    tag: str = field(compare=False)
    date: datetime.datetime = field(compare=True)
    commit_hash: str = field(compare=False)


@dataclass
class Organization:
    """See http//schema.org/Organization"""

    _id: str
    name: str
    legal_name: Optional[str] = None
    email: Optional[List[str]] = None
    description: Optional[str] = None
    logo: Optional[str] = None


class OrganizationSchema(JsonLDSchema):
    _id = fields.Id()
    name = fields.String(SDO.name)
    legal_name = fields.String(SDO.legalName)
    email = fields.String(SDO.email)
    description = fields.String(SDO.description)
    logo = fields.IRI(SDO.logo)

    class Meta:
        rdf_type = SDO.Organization
        model = Organization


@dataclass
class Person:
    """See http//schema.org/Person"""

    _id: str
    identifier: str
    name: Optional[str] = None
    email: Optional[str] = None
    affiliations: Optional[List[Organization]] = None

    def __str__(self):
        name = f"({self.name}) " if self.name else ""
        email = f"<{self.email}> " if self.email else ""
        orgs = (
            f"[{', '.join([org.name for org in self.affiliations])}]"
            if self.affiliations
            else ""
        )
        return f"{self.identifier} {name}{email}{orgs}".strip(" ")


class PersonSchema(JsonLDSchema):
    _id = fields.Id()
    identifier = fields.String(SDO.identifier)
    name = fields.String(SDO.name)
    affiliations = fields.Nested(
        SDO.affiliation, OrganizationSchema, many=True
    )

    class Meta:
        rdf_type = SDO.Person
        model = Person



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract classes for gimie objects."""
from abc import ABC, abstractmethod
from typing import Optional


from rdflib import Graph


class Extractor(ABC):
    """Extractor is an Abstract Base Class. It is only meant
    to define a standard interface for all extractors.

    All subclasses must implement extract() and to_graph() methods
    they are free to override the default serialize() and jsonld()
    """

    def __init__(self, path: str, _id: Optional[str] = None):
        self.path = path
        if _id is not None:
            self._id = _id

    @abstractmethod
    def extract(self):
        """Extract metadata"""
        ...

    @abstractmethod
    def to_graph(self) -> Graph:
        """Generate an RDF graph from the instance"""
        return Graph()

    def serialize(self, format: str = "ttl", **kwargs) -> str:
        """Serialize the RDF graph representing the instance."""
        return self.to_graph().serialize(format=format, **kwargs)  # type: ignore

    def jsonld(self) -> str:
        """Alias for jsonld serialization."""
        return self.serialize(format="json-ld")



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Orchestration of multiple extractors for a given project.
This is the main entry point for end-to-end analysis."""
from tempfile import gettempdir, TemporaryDirectory
from typing import Iterable, List, Optional, Union

import shutil
import git
from rdflib import Graph

from gimie.graph.operations import combine_graphs
from gimie.utils import validate_url
from gimie.sources import SOURCES
from gimie.sources.abstract import Extractor
from gimie.utils import validate_url


class Project:
    """A class to represent a project's git repository.


    Parameters
    ----------
    path :
        The path to the repository, either a file path or a URL.
    sources:
        The metadata sources to use.

    Examples
    --------
    >>> proj = Project("https://github.com/SDSC-ORD/gimie")
    """

    def __init__(
        self, path: str, sources: Optional[Union[str, Iterable[str]]] = None
    ):

        sources = normalize_sources(path, sources)
        # Remember if we cloned to cleanup at the end
        self._cloned = False
        if validate_url(path):
            self.url = path
            # We only need to clone a remote project
            # if a local extractor is enabled
            if any(map(is_local_source, sources)):
                self.project_dir = self.clone(path)
        else:
            self.url = None
            self.project_dir = path

        self.extractors = self.get_extractors(sources)
        for ex in self.extractors:
            ex.extract()

    def clone(self, url: str) -> str:
        """Clone target url in a new temporary directory"""
        target_dir = TemporaryDirectory().name
        cloned = git.Repo.clone_from(url, target_dir)  # type: ignore
        self._cloned = True

        return str(cloned.working_tree_dir)

    def get_extractors(self, sources: Iterable[str]) -> List[Extractor]:

        extractors: List[Extractor] = []
        for src in sources:
            if is_remote_source(src):
                extractor = get_extractor(self.url, src)  # type: ignore
            else:
                extractor = get_extractor(self.project_dir, src, _id=self.url)
            extractors.append(extractor)

        return extractors

    def to_graph(self) -> Graph:
        graphs = map(lambda ex: ex.to_graph(), self.extractors)
        combined_graph = combine_graphs(*graphs)
        return combined_graph

    def serialize(self, format: str = "ttl", **kwargs):
        return self.to_graph().serialize(format=format, **kwargs)

    def cleanup(self):
        """Recursively delete the project. Only works
        for remote (i.e. cloned) projects."""
        try:
            tempdir = gettempdir()
            in_temp = self.project_dir.startswith(tempdir)
            if self._cloned and in_temp:
                shutil.rmtree(self.project_dir)
        except AttributeError:
            pass

    def __del__(self):
        self.cleanup()


def normalize_sources(
    path: str, sources: Optional[Union[Iterable[str], str]] = None
) -> List[str]:
    """Input validation and normalization for metadata sources.
    Returns a list of all input sources.

    Parameters
    ----------
    path :
        The path to the repository, either a local path or a URL.
    sources:
        The metadata sources to use. If None only the git provider is used.

    Returns
    -------
    List[str]
        A list of all input sources. If the git provider was missing
        from input sources, it is inferred from path.

    Examples
    --------
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie")
    ['github']
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie", "github")
    ['github']
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie", ["github"])
    ['github']
    """
    if isinstance(sources, str):
        norm = [sources]
    elif isinstance(sources, Iterable):
        norm = list(sources)
    elif sources is None:
        norm = []
    else:
        raise ValueError(f"Invalid sources: {sources}")

    git_provider = get_git_provider(path)
    if git_provider not in norm:
        norm.append(git_provider)

    if (not validate_url(path)) and any(map(is_remote_source, norm)):
        raise ValueError("Cannot use a remote source with a local project.")
    return norm


def get_extractor(
    path: str, source: str, _id: Optional[str] = None
) -> Extractor:
    """Instantiate the correct extractor for a given source.

    Parameters
    -----------
    path
        Where the repository metadata is extracted from.
    source
        The source of the repository (git, gitlab, github, ...).
    _id
        The identifier to use for the repository. If not provided,
        it will be determined automatically by the extractor
    """
    return SOURCES[source].extractor(path, _id=_id)


def is_valid_source(source: str) -> bool:
    """Check if input is a valid source for gimie."""
    return source in SOURCES


def is_remote_source(source: str) -> bool:
    """Check if input is a valid remote source for gimie."""
    if is_valid_source(source):
        return SOURCES[source].remote
    return False


def is_local_source(source: str) -> bool:
    """Check if input is a valid local source for gimie."""
    return not is_remote_source(source)


def is_git_provider(source: str) -> bool:
    """Check if input is a valid git provider for gimie."""
    if is_valid_source(source):
        return SOURCES[source].git
    return False


def get_git_provider(url: str) -> str:
    """Given a git repository URL, return the corresponding git provider.
    Local path or unsupported git providers will return "git"."""
    # NOTE: We just check if the provider name is in the URL.
    # There may be a more robust way.
    if validate_url(url):
        for name, prov in SOURCES.items():
            if prov.git and prov.remote and name in url:
                return name
    # Fall back to local git if local path of unsupported provider
    return "git"



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import os
import requests
from typing import Any, Dict, List, Optional, Set, Union
from urllib.parse import urlparse
from dotenv import load_dotenv

from calamus import fields
from calamus.schema import JsonLDSchema
from rdflib import Graph

from gimie.sources.abstract import Extractor
from gimie.models import (
    Organization,
    OrganizationSchema,
    Person,
    PersonSchema,
)
from gimie.graph.namespaces import SDO
from gimie.sources.common.license import get_spdx_url
from gimie.sources.common.queries import (
    send_rest_query,
    send_graphql_query,
)

GH_API = "https://api.github.com"
load_dotenv()


def query_contributors(
    url: str, headers: Dict[str, str]
) -> List[Dict[str, Any]]:
    """Queries the list of contributors of target repository
    using Github's REST and GraphQL APIs. Returns a list of GraphQL User nodes.
    NOTE: This is a workaround for the lack of a contributors field in the GraphQL API."""
    owner, name = urlparse(url).path.strip("/").split("/")
    # Get contributors (available in the REST API but not GraphQL)
    data = f"repos/{owner}/{name}/contributors"
    contributors = send_rest_query(GH_API, data, headers=headers)
    ids = [contributor["node_id"] for contributor in contributors]
    # Get all contributors' metadata in 1 GraphQL query
    users_query = """
    query users($ids: [ID!]!) {
        nodes(ids: $ids) {
            ... on User {
                avatarUrl
                company
                login
                name
                organizations(first: 100) {
                    nodes {
                        avatarUrl
                        description
                        login
                        name
                        url
                    }
                }
                url
            }
        }
    }"""

    contributors = send_graphql_query(
        GH_API, users_query, data={"ids": ids}, headers=headers
    )
    # Drop empty users (e.g. dependabot)
    return [user for user in contributors["data"]["nodes"] if user]


@dataclass
class GithubExtractor(Extractor):
    """Extractor for GitHub repositories. Uses the GitHub GraphQL API to
    extract metadata into linked data."""

    path: str
    _id: Optional[str] = None
    token: Optional[str] = None

    name: Optional[str] = None
    author: Optional[Union[Organization, Person]] = None
    contributors: Optional[List[Person]] = None
    prog_langs: Optional[List[str]] = None
    download_url: Optional[str] = None
    description: Optional[str] = None
    date_created: Optional[datetime] = None
    date_modified: Optional[datetime] = None
    keywords: Optional[List[str]] = None
    license: Optional[str] = None
    software_version: Optional[str] = None

    def to_graph(self) -> Graph:
        """Convert repository to RDF graph."""
        jd = GithubExtractorSchema().dumps(self)
        g: Graph = Graph().parse(format="json-ld", data=str(jd))
        g.bind("schema", SDO)
        return g

    def extract(self):
        """Extract metadata from target GitHub repository."""
        if self._id is None:
            self._id = self.path
        self.name = urlparse(self.path).path.strip("/")
        data = self._fetch_repo_data(self.path)
        self.author = self._get_author(data["owner"])
        self.contributors = self._fetch_contributors(self.path)
        self.description = data["description"]
        self.date_created = datetime.fromisoformat(data["createdAt"][:-1])
        self.date_modified = datetime.fromisoformat(data["updatedAt"][:-1])
        # If license is available, convert to standard SPDX URL
        if data["licenseInfo"] is not None:
            self.license = get_spdx_url(data["licenseInfo"]["spdxId"])
        if data["primaryLanguage"] is not None:
            self.prog_langs = [data["primaryLanguage"]["name"]]
        self.keywords = self._get_keywords(*data["repositoryTopics"]["nodes"])
        last_release = data["latestRelease"]
        if last_release is not None:
            self.version = last_release["name"]
            self.download_url = (
                f"{self.path}/archive/refs/tags/{self.version}.tar.gz"
            )

    def _fetch_repo_data(self, url: str) -> Dict[str, Any]:
        """Fetch repository metadata from GraphQL endpoint."""
        owner, name = urlparse(url).path.strip("/").split("/")
        data = {"owner": owner, "name": name}
        repo_query = """
        query repo($owner: String!, $name: String!) {
            repository(name: $name, owner: $owner) {
                createdAt
                description
                latestRelease {
                    name
                }
                licenseInfo {
                    spdxId
                }
                mentionableUsers(first: 100) {
                    nodes {
                        login
                        name
                        avatarUrl
                        company
                        organizations(first: 100) {
                            nodes {
                                avatarUrl
                                description
                                login
                                name
                                url
                            }
                        }
                        url
                    }
                }
                name
                owner {
                    avatarUrl
                    login
                    url
                    ... on User {
                        company
                        name
                        organizations(first: 100) {
                            nodes {
                                avatarUrl
                                description
                                login
                                name
                                url
                            }
                        }
                    }
                    ... on Organization {
                        name
                        description
                    }
                }
                primaryLanguage {
                    name
                }
                repositoryTopics(first: 10) {
                    nodes {
                        topic {
                            name
                        }
                    }
                }
                updatedAt
                url
            }
        }
        """
        response = send_graphql_query(
            GH_API, repo_query, data, self._set_auth()
        )
        if "errors" in response:
            raise ValueError(response["errors"])

        return response["data"]["repository"]

    def _fetch_contributors(self, url: str) -> List[Person]:
        """Queries the GitHub GraphQL API to extract contributors through the commit list.
        NOTE: This is a workaround for the lack of a contributors field in the GraphQL API."""
        headers = self._set_auth()
        contributors = []
        resp = query_contributors(url, headers)
        for user in resp:
            contributors.append(self._get_user(user))
        return list(contributors)

    def _set_auth(self) -> Any:
        """Set authentication headers for GitHub API requests."""
        try:
            if not self.token:
                self.token = os.environ.get("GITHUB_TOKEN")
                assert self.token
            headers = {"Authorization": f"token {self.token}"}

            login = requests.get(f"{GH_API}/user", headers=headers)
            assert login.json().get("login")
        except AssertionError:
            return {}
        else:
            return headers

    def _get_keywords(self, *nodes: Dict[str, Any]) -> List[str]:
        """Extract names from GraphQL topic nodes."""
        return [node["topic"]["name"] for node in nodes]

    def _get_organization(self, node: Dict[str, Any]) -> Organization:
        """Extract details from a GraphQL organization node."""
        return Organization(
            _id=node["url"],
            name=node["login"],
            description=node["description"],
            legal_name=node["name"],
            logo=node["avatarUrl"],
        )

    def _get_author(self, node: Dict[str, Any]) -> Union[Organization, Person]:
        """Given the GraphQL node for a repository owner,
        return the author as a Person or Organization object."""

        if "organizations" in node:
            return self._get_user(node)

        return self._get_organization(node)

    def _get_user(self, node: Dict[str, Any]) -> Person:
        """Extract details from a GraphQL user node."""
        # Get user's affiliations
        orgs = [
            self._get_organization(org)
            for org in node["organizations"]["nodes"]
        ]
        return Person(
            _id=node["url"],
            identifier=node["login"],
            name=node["name"],
            affiliations=orgs,
        )


class GithubExtractorSchema(JsonLDSchema):
    """This defines the schema used for json-ld serialization."""

    _id = fields.Id()
    name = fields.String(SDO.name)
    author = fields.Nested(SDO.author, [PersonSchema, OrganizationSchema])
    contributors = fields.Nested(SDO.contributor, PersonSchema, many=True)
    prog_langs = fields.List(SDO.programmingLanguage, fields.String)
    download_url = fields.Raw(SDO.downloadUrl)
    description = fields.String(SDO.description)
    date_created = fields.Date(SDO.dateCreated)
    date_modified = fields.Date(SDO.dateModified)
    license = fields.IRI(SDO.license)
    path = fields.IRI(SDO.CodeRepository)
    keywords = fields.List(SDO.keywords, fields.String)
    version = fields.String(SDO.version)

    class Meta:
        rdf_type = SDO.SoftwareSourceCode
        model = GithubExtractor
        add_value_types = False



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Operations on graphs."""
from functools import reduce

from calamus import fields
from rdflib import Graph


schema = fields.Namespace("http://schema.org/")


def combine_graphs(*graphs: Graph) -> Graph:
    """Combines an arbitrary number of input graphs
    into a single graph."""
    return reduce(lambda g1, g2: g1 | g2, graphs)
